import os
import subprocess
import logging

logger = logging.getLogger(__name__)

tag = "0.8.0a1"
git_directory = os.environ.get("GE_DEV_DIR")


def get_git_describe_string():
    return subprocess.check_output(["git", "describe"]).decode().strip()


def get_git_revision_hash():
    return subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode()


def get_git_revision_short_hash():
    return subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode()


if git_directory is not None:
    start_dir = os.getcwd()

    try:
        os.chdir(git_directory)
        __version__ = get_git_describe_string()
    except subprocess.CalledProcessError:
        logger.warning("Unable to identify version tag using git.")
        __version__ = tag
    finally:
        os.chdir(start_dir)
else:
    logger.debug("Using default version tag.")
    __version__ = tag




import pytest

import os

from freezegun import freeze_time

try:
    from unittest import mock
except ImportError:
    import mock

from great_expectations.data_context.util import safe_mmkdir
from great_expectations.datasource.types import *
from great_expectations.datasource.generator import SubdirReaderGenerator, GlobReaderGenerator



@freeze_time("1955-11-05")
def test_batch_kwargs_fingerprint():
    test_batch_kwargs = PathBatchKwargs(
        {
            "path": "/data/test.csv"
        }
    )

    #demonstrate *output* kwargs post-datasource/generator

    # When there is only a single "important" key used in batch_kwargs, the ID can prominently include it
    assert test_batch_kwargs.batch_fingerprint == BatchFingerprint(
        partition_id="19551105T000000.000000Z",
        fingerprint="path:/data/test.csv")

    test_batch_kwargs = PathBatchKwargs(
        {
            "path": "/data/test.csv",
            "partition_id": "20190101"
        }
    )

    # When partition_id is explicitly included, we can extract it and potentially still have a human readable id
    assert test_batch_kwargs.batch_fingerprint == BatchFingerprint(
        partition_id="20190101",
        fingerprint="path:/data/test.csv")

    test_batch_kwargs = PathBatchKwargs(
        {
            "path": "/data/test.csv",
            "iterator": True,
            "partition_id": "3",
            "chunksize": 2e7,
            "parse_dates": [0, 3],
            "names": ["start", "type", "quantity", "end"]
        }
    )
    # When there are multiple relevant keys we use the hash of the batch_kwargs dictionary
    assert test_batch_kwargs.batch_fingerprint == BatchFingerprint(
        partition_id="3",
        fingerprint="a5d67721928ee13317a81459818a556b")


def test_batch_kwargs_from_dict():
    test_batch_kwargs = {
            "path": "/data/test.csv",
            "partition_id": "1"
        }

    # The build_batch_fingerprint convenience method makes it possible to build a batch_fingerprint from a dict.
    # HOWEVER, using it can be difficult since the default-ignored keys may depend on a specific batch_kwargs type
    assert BatchKwargs.build_batch_fingerprint(test_batch_kwargs) == BatchFingerprint(
        partition_id="1",
        fingerprint="path:/data/test.csv")


def test_glob_reader_path_partitioning():
    test_asset_globs = {
        "test_asset": {
            "glob": "*",
            "partition_regex": r"^((19|20)\d\d[- /.]?(0[1-9]|1[012])[- /.]?(0[1-9]|[12][0-9]|3[01]))_(.*)\.csv",
            "match_group_id": 1
        }
    }
    glob_generator = GlobReaderGenerator("test_generator", asset_globs=test_asset_globs)

    with mock.patch("glob.glob") as mock_glob:
        mock_glob_match = [
            "20190101__my_data.csv",
            "20190102__my_data.csv",
            "20190103__my_data.csv",
            "20190104__my_data.csv",
            "20190105__my_data.csv"
        ]
        mock_glob.return_value = mock_glob_match
        kwargs = [kwargs for kwargs in glob_generator.get_iterator("test_asset")]

    # GlobReaderGenerator uses partition_regex to extract partitions from filenames
    assert len(kwargs) == len(mock_glob_match)
    assert kwargs[0]["path"] == "20190101__my_data.csv"
    assert kwargs[0]["partition_id"] == "20190101"
    assert "timestamp" in kwargs[0]
    assert len(kwargs[0].keys()) == 3


def test_subdir_reader_path_partitioning(tmp_path_factory):
    base_directory = str(tmp_path_factory.mktemp("test_folder_connection_path"))
    mock_files = [
        "asset_1/20190101__asset_1.csv",
        "asset_1/20190102__asset_1.csv",
        "asset_1/20190103__asset_1.csv",
        "asset_2/20190101__asset_2.csv",
        "asset_2/20190102__asset_2.csv"
    ]
    for file in mock_files:
        safe_mmkdir(os.path.join(base_directory, file.split("/")[0]))
        open(os.path.join(base_directory, file), "w").close()

    subdir_reader_generator = SubdirReaderGenerator("test_generator", base_directory=base_directory)

    asset_1_kwargs = [kwargs for kwargs in subdir_reader_generator.get_iterator("asset_1")]
    asset_2_kwargs = [kwargs for kwargs in subdir_reader_generator.get_iterator("asset_2")]
    with pytest.raises(IOError):
        not_an_asset_kwargs = [kwargs for kwargs in subdir_reader_generator.get_iterator("not_an_asset")]

    assert len(asset_1_kwargs) == 3
    paths = set([kwargs["path"] for kwargs in asset_1_kwargs])
    assert paths == {
        os.path.join(base_directory, "asset_1/20190101__asset_1.csv"),
        os.path.join(base_directory, "asset_1/20190102__asset_1.csv"),
        os.path.join(base_directory, "asset_1/20190103__asset_1.csv")
    }
    partitions = set([kwargs["partition_id"] for kwargs in asset_1_kwargs])

    # SubdirReaderGenerator uses filenames from subdirectories to generate partition names
    assert partitions == {
        "20190101__asset_1",
        "20190102__asset_1",
        "20190103__asset_1"
    }
    assert "timestamp" in asset_1_kwargs[0]
    assert len(asset_1_kwargs[0].keys()) == 3

    # FIXME: We need to resolve whether this test should pass (because timestamp is *not* part of id)
    # assert asset_1_kwargs[0].batch_id == \
    #    asset_1_kwargs[0]["partition_id"] + "::" + "path:" + asset_1_kwargs[0]["path"]
    # FIXME: Or whether this one should pass (because timestamp *is* part of id)
    assert asset_1_kwargs[0].batch_fingerprint == BatchKwargs.build_batch_fingerprint(asset_1_kwargs[0])

    assert len(asset_2_kwargs) == 2
    paths = set([kwargs["path"] for kwargs in asset_2_kwargs])
    assert paths == {
        os.path.join(base_directory, "asset_2/20190101__asset_2.csv"),
        os.path.join(base_directory, "asset_2/20190102__asset_2.csv")
    }
    partitions = set([kwargs["partition_id"] for kwargs in asset_2_kwargs])
    assert partitions == {
        "20190101__asset_2",
        "20190102__asset_2"
    }
    assert "timestamp" in asset_2_kwargs[0]
    assert len(asset_2_kwargs[0].keys()) == 3



from six import string_types

from great_expectations.types import Config
#
#
# class DataContextConfig(Config):
#     _allowed_keys = set([
#         "ge_config_version",
#         "result_callback",
#         "config_variables_file_path",
#         "plugins_directory",
#         "expectations_store",
#         "profiling_store_name",
#         "evaluation_parameter_store_name",
#         "datasources",
#         "stores",
#         "data_docs",  # TODO: Rename this to sites, to remove a layer of extraneous nesting
#         "validation_operators",
#     ])
#
#     _required_keys = set([
#         # TODO next version re-introduce ge_config_version as required
#         # "ge_config_version",
#         "plugins_directory",
#         "expectations_store",
#         "profiling_store_name",
#         "evaluation_parameter_store_name",
#         "datasources",
#         "stores",
#         "data_docs",
#         # "validation_operators", # TODO: Activate!
#     ])
#
#     _key_types = {
#         "ge_config_version": int,
#         "config_variables_file_path": string_types,
#         "plugins_directory": string_types,
#         "expectations_store": dict,
#         "profiling_store_name": string_types,
#         "evaluation_parameter_store_name": string_types,
#         "datasources": dict,
#         "stores": dict,
#         "data_docs": dict,
#         "validation_operators": dict,
#     }



import os


class GreatExpectationsError(Exception):
    def __init__(self, message):
        self.message = message  


class DataContextError(GreatExpectationsError):
    pass


class InvalidConfigurationYamlError(GreatExpectationsError):
    pass


class InvalidTopLevelConfigKeyError(GreatExpectationsError):
    pass


class MissingTopLevelConfigKeyError(GreatExpectationsError):
    pass


class InvalidConfigValueTypeError(GreatExpectationsError):
    pass


class InvalidConfigVersionError(GreatExpectationsError):
    pass


class UnsupportedConfigVersionError(GreatExpectationsError):
    pass


class ZeroDotSevenConfigVersionError(GreatExpectationsError):
    pass


class ProfilerError(GreatExpectationsError):
    pass


class InvalidConfigError(DataContextError):
    def __init__(self, message):
        self.message = message
    
class ConfigNotFoundError(DataContextError):
    """
    the config file or the environments file cannot be found
    """
    def __init__(self, message):
        self.message = message

class ExpectationSuiteNotFoundError(GreatExpectationsError):
    def __init__(self, data_asset_name):
        self.data_asset_name = data_asset_name
        self.message = "No expectation suite found for data_asset_name %s" % data_asset_name

class BatchKwargsError(DataContextError):
    def __init__(self, message, batch_kwargs):
        self.message = message
        self.batch_kwargs = batch_kwargs

class DatasourceInitializationError(GreatExpectationsError):
    def __init__(self, datasource_name, message):
        self.message = "Cannot initialize datasource %s, error: %s" % (datasource_name, message)



# -*- coding: utf-8 -*-

PROJECT_HELP_COMMENT = """
# Welcome to great expectations. 
# This project configuration file allows you to define datasources, 
# generators, integrations, and other configuration artifacts that
# make it easier to use Great Expectations.

# For more help configuring great expectations, 
# see the documentation at: https://docs.greatexpectations.io/en/latest/core_concepts/data_context.html#configuration

# NOTE: GE uses the names of configured datasources and generators to manage
# how expectations and other configuration artifacts are stored in the 
# expectations/ and datasources/ folders. If you need to rename an existing
# datasource or generator, be sure to also update the paths for related artifacts.

ge_config_version: 1

"""

CONFIG_VARIABLES_INTRO = """
# Great Expectations config file supports variable substitution.

# Variable substitution enables these two use cases:
# 1. do not store sensitive credentials in a committed file (since the credentials file should be in 
#    the uncommitted directory  
# 2. allow a config parameter to take different values, depending on the environment (e.g., dev/staging/prod)
#
# When GE encounters the following syntax in the config file:
#
# my_key: ${my_value} (or $my_value)
#
# GE will attempt to replace the value of “my_key” with the value of env variable “my_value” or with the value of the key “my_value” read from credentials file (env variable takes precedence).
#
# If the replacing value comes from the config variables file, it can be a simple value or a dictionary. If it comes from an environment variable, it must be a simple value.
"""

PROJECT_OPTIONAL_CONFIG_COMMENT = CONFIG_VARIABLES_INTRO + """
config_variables_file_path: uncommitted/config_variables.yml

# The plugins_directory is where the data_context will look for custom_data_assets.py
# and any configured evaluation parameter store

plugins_directory: plugins/

expectations_store:
  class_name: ExpectationStore
  store_backend:
    class_name: FixedLengthTupleFilesystemStoreBackend
    base_directory: expectations/

profiling_store_name: local_validation_result_store

evaluation_parameter_store_name: evaluation_parameter_store

# Configure additional data context options here.

# Uncomment the lines below to enable s3 as a result store. If a result store is enabled,
# validation results will be saved in the store according to run id.

# For S3, ensure that appropriate credentials or assume_role permissions are set where
# validation happens.

stores:

  local_validation_result_store:
    class_name: ValidationResultStore
    store_backend:
      class_name: FixedLengthTupleFilesystemStoreBackend
      base_directory: uncommitted/validations/
      filepath_template: '{4}/{0}/{1}/{2}/{3}.json'

  # s3_validation_result_store:
  #   class_name: ValidationStore
  #   store_backend:
  #     class_name: FixedLengthTupleS3StoreBackend
  #     bucket: ???
  #     prefix: ???
  #     file_extension: json
  #     filepath_template: '{4}/{0}/{1}/{2}/{3}.{file_extension}'

  # FIXME: These configs are temporarily commented out to facititate refactoring Stores.

  # local_profiling_store:
  #   module_name: great_expectations.data_context.store
  #   class_name: FilesystemStore
  #   store_config:
  #     base_directory: uncommitted/profiling/
  #     serialization_type: json
  #     file_extension: .json

  # local_workbench_site_store:
  #   module_name: great_expectations.data_context.store
  #   class_name: FilesystemStore
  #   store_config:
  #     base_directory: uncommitted/documentation/local_site
  #     file_extension: .html

  # shared_team_site_store:
  #   module_name: great_expectations.data_context.store
  #   class_name: FilesystemStore
  #   store_config:
  #     base_directory: uncommitted/documentation/team_site
  #     file_extension: .html

  # fixture_validation_results_store:
  #   module_name: great_expectations.data_context.store
  #   class_name: FilesystemStore
  #   store_config:
  #     base_directory: fixtures/validations
  #     file_extension: .zzz

  fixture_validation_results_store:
    class_name: ValidationResultStore
    store_backend:
      class_name: FixedLengthTupleFilesystemStoreBackend
      base_directory: fixtures/validations
      filepath_template: '{4}/{0}/{1}/{2}/{3}.json'

#  data_asset_snapshot_store:
#    module_name: great_expectations.data_context.store
#    class_name: S3Store
#    store_config:
#      bucket:
#      key_prefix:

  evaluation_parameter_store:
    module_name: great_expectations.data_context.store
    class_name: EvaluationParameterStore

  local_site_html_store:
    module_name: great_expectations.data_context.store
    class_name: HtmlSiteStore
    base_directory: uncommitted/documentation/local_site/

  team_site_html_store:
    module_name: great_expectations.data_context.store
    class_name: HtmlSiteStore
    base_directory: uncommitted/documentation/team_site/


validation_operators:
  # Read about validation operators at: https://docs.greatexpectations.io/en/latest/guides/validation_operators.html
  perform_action_list_operator:
    class_name: PerformActionListValidationOperator
    action_list:
      - name: store_validation_result
        action:
          class_name: StoreAction
          target_store_name: local_validation_result_store
      - name: store_evaluation_params
        action:
          class_name: ExtractAndStoreEvaluationParamsAction
          target_store_name: evaluation_parameter_store
      - name: store_evaluation_params
        action:
          class_name: SlackNotificationAction
          slack_webhook: ${validation_notification_slack_webhook}
#          notify_on: all
          renderer:
            module_name: great_expectations.render.renderer.slack_renderer
            class_name: SlackRenderer
    
# Uncomment the lines below to enable a result callback.

# result_callback:
#   slack: ${slack_callback_url}

# TODO : Remove the extra layer of yml nesting in v0.8:
data_docs:
  sites:
    local_site: # site name
    # “local_site” renders documentation for all the datasources in the project from GE artifacts in the local repo. 
    # The site includes expectation suites and profiling and validation results from uncommitted directory. 
    # Local site provides the convenience of visualizing all the entities stored in JSON files as HTML.

      # specify a whitelist here if you would like to restrict the datasources to document
      datasource_whitelist: '*'

      module_name: great_expectations.render.renderer.site_builder
      class_name: SiteBuilder
      target_store_name: local_site_html_store
      
      site_index_builder:
        class_name: DefaultSiteIndexBuilder
      
      site_section_builders:
          
        expectations:
          class_name: DefaultSiteSectionBuilder
          source_store_name: expectations_store
          renderer:
            module_name: great_expectations.render.renderer
            class_name: ExpectationSuitePageRenderer

        validations:
          class_name: DefaultSiteSectionBuilder
          source_store_name: local_validation_result_store
          run_id_filter:
            ne: profiling
          renderer:
            module_name: great_expectations.render.renderer
            class_name: ValidationResultsPageRenderer

        profiling:
          class_name: DefaultSiteSectionBuilder
          source_store_name: local_validation_result_store
          run_id_filter:
            eq: profiling
          renderer:
            module_name: great_expectations.render.renderer
            class_name: ProfilingResultsPageRenderer

    team_site:
    # "team_site" is meant to support the "shared source of truth for a team" use case. 
    # By default only the expectations section is enabled.
    #  Users have to configure the profiling and the validations sections (and the corresponding validations_store and profiling_store attributes based on the team's decisions where these are stored (a local filesystem or S3). 
    # Reach out on Slack (https://greatexpectations.io/slack>) if you would like to discuss the best way to configure a team site.

      # specify a whitelist here if you would like to restrict the datasources to document
      datasource_whitelist: '*'
      
      module_name: great_expectations.render.renderer.site_builder
      class_name: SiteBuilder
      target_store_name: team_site_html_store
      
      site_index_builder:
        class_name: DefaultSiteIndexBuilder
      
      site_section_builders:
          
        expectations:
          class_name: DefaultSiteSectionBuilder
          source_store_name: expectations_store
          renderer:
            module_name: great_expectations.render.renderer
            class_name: ExpectationSuitePageRenderer

"""

PROJECT_TEMPLATE = PROJECT_HELP_COMMENT + "datasources: {}\n" + PROJECT_OPTIONAL_CONFIG_COMMENT

CONFIG_VARIABLES_COMMENT = CONFIG_VARIABLES_INTRO

CONFIG_VARIABLES_FILE_TEMPLATE = CONFIG_VARIABLES_INTRO